/*
 * PINsafeClient.java
 */

package com.swiveltechnologies.client;

import java.io.*;
import java.net.*;
import java.util.Properties;
import java.util.ArrayList;
import java.lang.Character;

/**
 * Provides a convenient object for accessing the authentication services of a
 * PINsafe Server.
 *
 * @author Chris Noble, James Stembridge
 * @version $Id: PINsafeClient.java 2168 2007-10-05 14:35:17Z chris $
 */

public class PINsafeClient {
	public static Character	 [] bannedCharacters = {0x00, 0x0a, 0x0d,0x1b,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x3b,0x3c,0x3d,0x3e,0x60,0x7c,0x7e};
    //Member field variables for getters and setters...
    private boolean _ssl     = false;
    private String  _server  = "localhost";
    private String  _port    = "8080";
    private String  _context = "/pinsafe";
    private String  _secret  = "";
    private String  _error   = "";
    private ArrayList<String>  _warning = null;

    private String elementName;
    private int index =0;

    private String _request;
    private String _response;

    

    /**
     * Creates a new instance of PINsafeClient.
     */
    public PINsafeClient() {
    }

    /**
     * Creates a new instance of PINsafeClient with the provided properties.
     * @param props Properties of the PINsafe server. Values are taken
     * from the "ssl", "server", "port", "context" and "secret" keys.
     */
    public PINsafeClient(Properties props) {
        _ssl     = Boolean.valueOf(props.getProperty("ssl", String.valueOf(_ssl)));
        _server  = props.getProperty("server",  _server);
        _port    = props.getProperty("port",    _port);
        _context = props.getProperty("context", _context);
        _secret  = props.getProperty("secret",  _secret);
    }

   public static boolean checkChars(String s){
	   String t;
	   for(int c=0; c<bannedCharacters.length; c++){
		   t = bannedCharacters[c].toString();
		   if(s.contains(t))
			   return false;
	   }
		return true;   
   }
    
    /**
     * Gets the XML response from the PINsafe server.
     * @return The last XML response from the PINsafe server.
     */
    public String getResponse() {
        if (_response == null) {
            return "";
        } else {
            return _response.toString();
        }
    }

    /**
     * Gets the XML request sent to the PINsafe Server
     * @return The last XML request sent to the PINsafe Server.
     */
    public String getRequest() {
        if (_request == null) {
            return "";
        } else {
            return _request.toString();
        }
    }

    /**
     * Gets the current error message if there is one.
     * @return Last error message returned from the PINsafe Server.
     */
    public String getError() {
        return _error;
    }
    /**
     * Extracts any warnings from the PINsafe response
     * @return Last warning messages returned from the PINsafe Server..
     */
    private ArrayList <String> extractWarnings() {
        ArrayList<String>tempwarning = new ArrayList<String>();
        String warning = extractElement("Warning");
        while(warning != ""){
        	tempwarning.add(warning);
        	warning = extractElement("Warning");
        }
    	return tempwarning;
    }

    /**
     * Returns the security string in plain text
     * If an error occurs an empty string will be returned
     * @param
     * @return String securityString, Session ID for the session, or an empty string if no session could
     * be start
     */
    public String getSecurityString(){
    	String securityString = "";
    	securityString = extractElement("SecurityString");
    	return securityString;
    }

    /**
     * Returns the security string in plain text, post 3.2
     * If an error occurs an empty string will be returned
     * @param
     * @return String securityString, Session ID for the session, or an empty string if no session could
     * be start
     */
    public String getSecurityString(String sessionID){
    	String securityString = "";
    	StringBuffer b = new StringBuffer();

		//
		char[] buff = new char[512];
		try{
		BufferedReader r = new BufferedReader(new InputStreamReader(singleChannelText(sessionID)));
		for (int n; (n = r.read(buff)) > 0;) {
			//
			b.append(buff, 0, n);
		}
		r.close();

		//
		securityString = b.toString();
		if(securityString.length() > 10){
			securityString = securityString.split("\n")[1];
		} else {
			securityString = securityString.split("\n")[0];
		}
		} catch(Exception e){
			return("");
		}
    	return securityString;
    }

    /**
     * Returns the security string in plain text, post 3.2
     * If an error occurs an empty string will be returned
     * @param
     * @return String securityString, Session ID for the session, or an empty string if no session could
     * be start
     */
    public String ping(){
        StringBuffer tmp = new StringBuffer();

        tmp.append("<?xml version=\"1.0\" ?>");
        tmp.append("<SASRequest>");
        tmp.append("<Version>3.1</Version>");
        tmp.append("<Action>ping</Action>");
        tmp.append("</SASRequest>");

        _request = tmp.toString();

        //Send request for processing by PINsafe server...
        if (!this.sendRequest()) {
            return "";
        }

        //See if we received a pass from the server...
        if (extractElement("Result").equalsIgnoreCase("PASS")) {
            return ("pass");
        } else {
            _error = extractElement("Error");
            return "";
        }

    }




    /**
     * Gets the current warning messages if there are any.
     * @return Last warning messages returned from the PINsafe Server.
     */
    public ArrayList<String> getWarnings() {
        return _warning;
    }

    /**
     * Sets whether SSL is used for communicating with the PINsafe server.
     * @param value A <code>boolean</code> inidicating if SSL should be used.
     */
    public void setSSL(boolean value) {
        _ssl = value;
    }

    /**
     * Gets whether SSL is used for communicating with the PINsafe server.
     * @return A <code>boolean</code> indicating if SSL is used.
     */
    public boolean getSLL() {
        return _ssl;
    }

    /**
     * Sets the server hostname or IP address.
     * @param server The hostname or IP address of the PINsafe server.
     */
    public void setServer(String server) {
        _server = server;
    }

    /**
     * Gets the current hostname or IP address of the PINsafe server.
     * @return The current hostname or IP address of the PINsafe server.
     */
    public String getServer() {
        return _server;
    }

    /**
     * Sets the server port number.
     * @param port The port number of the PINsafe server.
     */
    public void setPort(int port) {
        _port = String.valueOf(port);
    }

    /**
     * Gets the current port number of the PINsafe server.
     * @return The current port number of the PINsafe server.
     */
    public int getPort() {
        return Integer.valueOf(_port).intValue();
    }

    /**
     * Sets the context used for communicating with the PINsafe server.
     * @param context The context used for communicating  with the PINsafe server.
     */
    public void setContext(String context) {
        _context = context;
    }

    /**
     * Gets the context used for commuinicating with the PINsafe server.
     * @return The current context used for commuinicating with the PINsafe server.
     */
    public String getContext() {
        return _context;
    }

    /**
     * Sets the secret used for communicating with the PINsafe server.
     * @param secret The secret used for communicating  with the PINsafe server.
     */
    public void setSecret(String secret) {
        _secret = secret;
    }

    /**
     * Gets the secret used for commuinicating with the PINsafe server.
     * @return The current secret used for commuinicating with the PINsafe server.
     */
    public String getSecret() {
        return _secret;
    }

    /**
     * Sends an XML request to the PINsafe server.
     * @param XML The XML request to be sent to the PINsafe server.
     * @return The XML response from the PINsafe server, or an empty string if
     * an error occured.
     */
    public String sendRequest(String XML) {
        try {
            _request = XML;
            sendRequest();
            return _response;
        } catch (Exception e) {
            _error = e.getLocalizedMessage();
            return "";
        }
    }

    /**
     * Sends the request present in _request to the PINsafe server for processing.
     * Any XML response from the server will be returned to the caller.  If there
     * is an error an empty string will be returned and further details can be
     * obtained from the Error property.
     * @return Result from the PINsafe server, an empty string if there is a problem.
     */
    private boolean sendRequest() {
        try {
            //Create a URL object using the current URL...
            StringBuffer address = new StringBuffer();
            if (_ssl) {
                address.append("https://");
            } else {
                address.append("http://");
            }
            address.append(_server + ":" + _port + "/" + _context + "/AgentXML");

            URL url = new URL(address.toString());

            //Open connection
            HttpURLConnection conn = (HttpURLConnection)url.openConnection();
            conn.setDoInput(true);
            conn.setDoOutput(true);
            conn.setUseCaches(false);
            conn.setRequestMethod("POST");
            conn.setRequestProperty("Content-Type", "application/x-www-form-urlencoded");
            conn.setRequestProperty("Content-Encoding", "UTF-8");

            //Write request to connection...
            DataOutputStream out = new DataOutputStream(conn.getOutputStream());
            out.writeBytes(URLEncoder.encode(_request, "UTF8"));

            out.flush();
            out.close();

            //Check if response was 200 OK...
            int responseCode = conn.getResponseCode();
            if (responseCode == HttpURLConnection.HTTP_OK) {
                //Get Response...
                BufferedReader br = new BufferedReader(
                		new InputStreamReader(conn.getInputStream()));
                StringBuffer tmp = new StringBuffer();
                String oneline;

                while ( (oneline = br.readLine()) != null) {
                    tmp.append(oneline);
                }
                br.close();

                _response = tmp.toString();

                return true;
            } else {
                _error = responseCode + ": " + conn.getResponseMessage();
                return false;
            }
        } catch (Exception e) {
            _error = e.getLocalizedMessage();
            return false;
        }
    }

    /**
     * Extracts an element value from the response from the server.
     * If the call is repeated with the same element, the next instance is returned.
     * @param element The names of the element we want to extract the value for.
     * @return The value of the element named in the element parameter.
     */
    private String extractElement(String element) {

    	try {
    		if(!element.equals(elementName)){
    			elementName = element;
    			index = 0;
    		}

            //Construct a start and end element...
            String startElement = "<" + element + ">";
            String endElement = "</" + element + ">";

            //Find the start and end points for the element...
            int start = _response.indexOf(startElement, index);
            int end = _response.indexOf(endElement, start);

            //Check there is an element value from the server...
            if ((start == -1) || (end == -1)) {
                return "";
            }

            //Shuffle along to the start of the error message...
            start += startElement.length();
            index = end;
            //Extract the error message and display it...

            return _response.substring(start, end);

        } catch(Exception ex) {
            _error = ex.getLocalizedMessage();
            return "";
        }
    }

    /**
     * Starts a new session with the PINsafe server.
     * If an error occurs an empty string will be returned and a detailed error
     * message can be obtained using the <code>getError</code> method.
     * @param username The username for the user.
     * @return Session ID for the session, or an empty string if no session could
     * be started.
     */
    public String startSession(String name){
    	return (startSession(name,""));
    }
    public String startSession(String username, String sessionContext) {
        //Construct the request for the Server...
        StringBuffer tmp = new StringBuffer();

        tmp.append("<?xml version=\"1.0\" ?>");
        tmp.append("<SASRequest>");
        tmp.append("<Version>3.1</Version>");
        tmp.append("<Action>sessionstart</Action>");
        tmp.append("<Username>");
        tmp.append(username);
        tmp.append("</Username>");
        if(sessionContext.length() > 0){
        	tmp.append("<sessionContext>");
        	tmp.append(sessionContext);
        	tmp.append("</sessionContext>");
        }
        tmp.append("<Secret>");
        tmp.append(_secret);
        tmp.append("</Secret>");
        tmp.append("</SASRequest>");

        _request = tmp.toString();

        //Send request for processing by PINsafe server...
        if (!this.sendRequest()) {
            return "";
        }

        //See if we received a pass from the server...
        if (extractElement("Result").equalsIgnoreCase("PASS")) {
            return extractElement("SessionID");
        } else {
            _error = extractElement("Error");
            return "";
        }
    }

    /**
     * Get a single channel image from the PINsafe server.
     * @param sessionID Session ID returned by <code>startSession</code>.
     * @return An <code>InputStream</code> for reading the image data, or null
     * if an error occured.
     */
    public InputStream singleChannelImage(String sessionID) {
    	try {
            //Create a URL object using the current URL...
            StringBuffer address = new StringBuffer();
            if (_ssl) {
                address.append("https://");
            } else {
                address.append("http://");
            }
            address.append(_server + ":" + _port + "/" + _context +
            		"/SCImage?sessionid=" + URLEncoder.encode(sessionID, "UTF8"));

            URL url = new URL(address.toString());

            //Open connection
            HttpURLConnection conn = (HttpURLConnection)url.openConnection();
            conn.setDoInput(true);
            conn.setDoOutput(false);
            conn.setUseCaches(false);
            conn.setRequestMethod("GET");

            //Check if response was 200 OK...
            int responseCode = conn.getResponseCode();
            if (responseCode == HttpURLConnection.HTTP_OK) {
            	//Return InputStream for reading image data
            	return conn.getInputStream();
            } else {
                _error = responseCode + ": " + conn.getResponseMessage();
                return null;
            }
    	} catch (Exception e) {
    		_error = e.getLocalizedMessage();
    		return null;
    	}
    }

    /**
     * Get a single channel secuirty string via text from the PINsafe server.
     * @param sessionID Session ID returned by <code>startSession</code>.
     * @return An <code>InputStream</code> for reading the image data, or null
     * if an error occured.
     */
    public InputStream singleChannelText(String sessionID) {
    	try {
            //Create a URL object using the current URL...
            StringBuffer address = new StringBuffer();
            if (_ssl) {
                address.append("https://");
            } else {
                address.append("http://");
            }
            address.append(_server + ":" + _port + "/" + _context +
            		"/SCText?sessionid=" + URLEncoder.encode(sessionID, "UTF8"));

            URL url = new URL(address.toString());

            //Open connection
            HttpURLConnection conn = (HttpURLConnection)url.openConnection();
            conn.setDoInput(true);
            conn.setDoOutput(false);
            conn.setUseCaches(false);
            conn.setRequestMethod("GET");

            //Check if response was 200 OK...
            int responseCode = conn.getResponseCode();
            if (responseCode == HttpURLConnection.HTTP_OK) {
            	//Return InputStream for reading image data
            	return conn.getInputStream();
            } else {
                _error = responseCode + ": " + conn.getResponseMessage();
                return null;
            }
    	} catch (Exception e) {
    		_error = e.getLocalizedMessage();
    		return null;
    	}
    }

    /**
     * Request that a dual channel message be sent by the PINsafe server.
     * Note that on-demand dual channel message sending must be enabled on
     * the server for this request to result in a message being sent.
     * @param sessionID Session ID returned by <code>startSession</code>.
     * @return A <code>boolean</code> indicating if the request was
     * successful.
     */
    public boolean dualChannelMessage(String sessionID) {
    	try {
            //Create a URL object using the current URL...
            StringBuffer address = new StringBuffer();
            if (_ssl) {
                address.append("https://");
            } else {
                address.append("http://");
            }
            address.append(_server + ":" + _port + "/" + _context +
            		"/DCMessage?sessionid=" + URLEncoder.encode(sessionID, "UTF8"));

            URL url = new URL(address.toString());

            //Open connection
            HttpURLConnection conn = (HttpURLConnection)url.openConnection();
            conn.setDoInput(true);
            conn.setDoOutput(false);
            conn.setUseCaches(false);
            conn.setRequestMethod("GET");

            //Check if response was 200 OK...
            int responseCode = conn.getResponseCode();
            if (responseCode == HttpURLConnection.HTTP_OK) {
            	//Assume request was successful
            	return true;
            } else {
                _error = responseCode + ": " + conn.getResponseMessage();
                return false;
            }
    	} catch (Exception e) {
    		_error = e.getLocalizedMessage();
    		return false;
    	}
    }

    /**
     * Attempts to perform a login with the PINsafe server.  Any error returned
     * by the PINsafe server can be retrieved using the <code>getError</code> method.
     * @param username The username of the user.
     * @param password The password for the user.
     * @param OTC The One-Time-Code for the user.
     * @return Boolean value indicating if the login was successful or not.
     */
    public boolean login(String username, String password, String OTC) {
        //Passwords might not be in use
        if (password == null) {
            password = "";
        }

        //Construct the request for the Server...
        StringBuffer tmp = new StringBuffer();

        tmp.append("<?xml version=\"1.0\" ?>");
        tmp.append("<SASRequest>");
        tmp.append("<Version>3.1</Version>");
        tmp.append("<Action>login</Action>");
        tmp.append("<Secret>");
        tmp.append(_secret);
        tmp.append("</Secret>");
        tmp.append("<Username>");
        tmp.append(username);
        tmp.append("</Username>");
        tmp.append("<Password>");
        tmp.append(password);
        tmp.append("</Password>");
        tmp.append("<OTC>");
        tmp.append(OTC);
        tmp.append("</OTC>");
        tmp.append("</SASRequest>");

        _request = tmp.toString();

        //Send request for processing by PINsafe server...
        if (!this.sendRequest()) {
            return false;
        }
        //See if we received a pass from the server...
        if (extractElement("Result").equalsIgnoreCase("PASS")) {
        _warning = extractWarnings();
        	return true;
        } else {
            _error = extractElement("Error");
            return false;
        }
    }

    /**
     * Attempts to perform a change PIN with the PINsafe Server.  Any error returned
     * by the PINsafe server can be retrieved using the <code>getError</code> method.
     * @param username The username of the user.
     * @param password The password for the user.
     * @param newPassword The new password for the user.
     * @param OTC The One-Time-Code for the user.
     * @param newOTC The new One-Time-Code for the user.
     * @return Boolean value indicating if the PIN change was successful or not.
     */
    public boolean changePIN(String username, String password, String newPassword,
    		String OTC, String newOTC) {

        //Handle the case when passwords aren't in use
        if (password == null) {
            password = "";
        }
        if (newPassword == null) {
            newPassword = "";
        }

        //Construct the request for the Server...
        StringBuffer tmp = new StringBuffer();

        tmp.append("<?xml version=\"1.0\" ?>");
        tmp.append("<SASRequest>");
        tmp.append("<Version>3.1</Version>");
        tmp.append("<Action>changepin</Action>");
        tmp.append("<Secret>");
        tmp.append(_secret);
        tmp.append("</Secret>");
        tmp.append("<Username>");
        tmp.append(username);
        tmp.append("</Username>");
        tmp.append("<Password>");
        tmp.append(password);
        tmp.append("</Password>");
        tmp.append("<NewPassword>");
        tmp.append(newPassword);
        tmp.append("</NewPassword>");
        tmp.append("<OTC>");
        tmp.append(OTC);
        tmp.append("</OTC>");
        tmp.append("<NewOTC>");
        tmp.append(newOTC);
        tmp.append("</NewOTC>");
        tmp.append("</SASRequest>");

        _request = tmp.toString();

        //Send request for processing by PINsafe server...
        if (!this.sendRequest()) {
            return false;
        }

        //See if we received a pass from the server...
        if (extractElement("Result").equalsIgnoreCase("PASS")) {
            return true;
        } else {
            _error = extractElement("Error");
            return false;
        }
    }

    /**
     * Send a request for a reset code to the PINsafe server.  Any error returned
     * by the PINsafe server can be retrieved using the <code>getError</code> method.
     * @param username The username of the user.
     * @return Boolean value indicating if the request was successful or not.
     */
    public boolean resetCode(String username) {
        //Construct the request for the Server...
        StringBuffer tmp = new StringBuffer();

        tmp.append("<?xml version=\"1.0\" ?>");
        tmp.append("<SASRequest>");
        tmp.append("<Version>3.1</Version>");
        tmp.append("<Action>resetcode</Action>");
        tmp.append("<Secret>");
        tmp.append(_secret);
        tmp.append("</Secret>");
        tmp.append("<Username>");
        tmp.append(username);
        tmp.append("</Username>");
        tmp.append("</SASRequest>");

        _request = tmp.toString();

        //Send request for processing by PINsafe server...
        if (!this.sendRequest()) {
            return false;
        }

        //See if we received a pass from the server...
        if (extractElement("Result").equalsIgnoreCase("PASS")) {
            return true;
        } else {
            _error = extractElement("Error");
            return false;
        }
    }

    /**
     * Attempts to perform a user reset with the PINsafe Server.  Any error returned
     * by the PINsafe server can be retrieved using the <code>getError</code> method.
     * @param username The username of the user.
     * @param resetCode The reset code for the user.
     * @return Boolean value indicating if the PIN change was successful or not.
     */
    public boolean reset(String username, String resetCode) {
        //Construct the request for the Server...
        StringBuffer tmp = new StringBuffer();

        tmp.append("<?xml version=\"1.0\" ?>");
        tmp.append("<SASRequest>");
        tmp.append("<Version>3.1</Version>");
        tmp.append("<Action>reset</Action>");
        tmp.append("<Secret>");
        tmp.append(_secret);
        tmp.append("</Secret>");
        tmp.append("<Username>");
        tmp.append(username);
        tmp.append("</Username>");
        tmp.append("<ResetCode>");
        tmp.append(resetCode);
        tmp.append("</ResetCode>");
        tmp.append("</SASRequest>");

        _request = tmp.toString();

        //Send request for processing by PINsafe server...
        if (!this.sendRequest()) {
            return false;
        }

        //See if we received a pass from the server...
        if (extractElement("Result").equalsIgnoreCase("PASS")) {
            return true;
        } else {
            _error = extractElement("Error");
            return false;
        }
    }
    
    enum status {
    	init,
    	
    	sessionStart,
    	
    	changePin
    }
   }

